#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#define N 16

void init_array(double * array, int len){
    for (int i=0; i<len; i++){
        array[i] = (double) i;
    }
}

void print_array(double * array, int len){
    printf("[ ");
    for (int i=0; i<len; i++){
        printf("%lf ", array[i]);
    }
    printf("]\n");
}

void mpi_normalize(double * array, int len){

    int rank, comm_size;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &comm_size);

    int local_size = N / comm_size; //assume comm_size divides N
    double * local_array = (double *) malloc(sizeof(double)*local_size);

    MPI_Scatter(array, local_size, MPI_DOUBLE, local_array, local_size, MPI_DOUBLE, 0, MPI_COMM_WORLD);

    double length_squared=0, local_length_squared;
    for (int i=0; i<local_size; i++){
        local_length_squared += local_array[i] * local_array[i];
    }

    MPI_Allreduce(&local_length_squared, &length_squared, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

    double inv_length_squared = 1 / length_squared;

    for (int i=0; i<local_size; i++){
        local_array[i] *= inv_length_squared;
    }

    MPI_Gather(local_array, local_size, MPI_DOUBLE, array, local_size, MPI_DOUBLE, 0, MPI_COMM_WORLD);

    free(local_array);
}

/*
void seq_normalize(double * array, int len){
    double length_squared=0;
    for (int i=0; i<len; i++){
        length_squared += array[i] * array[i];
    }

    double inv_length_squared = 1 / length_squared;
        
    for (int i=0; i<len; i++){
        array[i] *= inv_length_squared;
    }
}
*/

int main(int argc, char * argv[]){

    double array[N];
    int rank, comm_size;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    if (rank==0){
        init_array(array, N);
        print_array(array, N);
    }
   
    mpi_normalize(array, N);

    if (rank==0){
        print_array(array, N);
    }

    MPI_Finalize();
    
    return 0;
}
